/* SPDX-License-Identifier: GPL-2.0-or-later */

/***************************************************************************
 *   Copyright (C) 2007 by Dominic Rath                                    *
 *   Dominic.Rath@gmx.de                                                   *
 ***************************************************************************/

#ifndef OPENOCD_TARGET_ETM_DUMMY_H
#define OPENOCD_TARGET_ETM_DUMMY_H

#include "etm.h"

extern struct etm_capture_driver etm_dummy_capture_driver;

#endif /* OPENOCD_TARGET_ETM_DUMMY_H */
