/* SPDX-License-Identifier: GPL-2.0-or-later */

/***************************************************************************
 *   Copyright (C) 2011 by Mathias Kuester                                 *
 *   Mathias Kuester <kesmtp@freenet.de>                                   *
 *                                                                         *
 *   Copyright (C) 2012 by Spencer Oliver                                  *
 *   spen@spen-soft.co.uk                                                  *
 ***************************************************************************/

#ifndef OPENOCD_JTAG_HLA_HLA_TRANSPORT_H
#define OPENOCD_JTAG_HLA_HLA_TRANSPORT_H

enum hl_transports {
	HL_TRANSPORT_UNKNOWN = 0,
	HL_TRANSPORT_SWD,
	HL_TRANSPORT_JTAG,
};

#endif /* OPENOCD_JTAG_HLA_HLA_TRANSPORT_H */
