/* SPDX-License-Identifier: GPL-2.0-or-later */

#ifndef OPENOCD_RTOS_RTOS_ECOS_STACKINGS_H
#define OPENOCD_RTOS_RTOS_ECOS_STACKINGS_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "rtos.h"

extern const struct rtos_register_stacking rtos_ecos_cortex_m3_stacking;

#endif /* OPENOCD_RTOS_RTOS_ECOS_STACKINGS_H */
