/* SPDX-License-Identifier: GPL-2.0-or-later */

/***************************************************************************
 *   Copyright (C) 2013-2014,2019-2020 Synopsys, Inc.                      *
 *   Frank Dols <frank.dols@synopsys.com>                                  *
 *   Mischa Jonker <mischa.jonker@synopsys.com>                            *
 *   Anton Kolesov <anton.kolesov@synopsys.com>                            *
 *   Evgeniy Didin <didin@synopsys.com>                                    *
 ***************************************************************************/

#ifndef OPENOCD_TARGET_ARC_CMD_H
#define OPENOCD_TARGET_ARC_CMD_H

extern const struct command_registration arc_monitor_command_handlers[];

#endif /* OPENOCD_TARGET_ARC_CMD_H */
